from types import MappingProxyType
from typing import Any, Dict, NamedTuple, Optional, Union

import rdflib

OCTA = rdflib.Namespace('https://ns.octadocs.io/')
LOCAL = rdflib.Namespace('local:')


# To prevent WPS401 Found mutable module constant
DEFAULT_NAMESPACES = MappingProxyType({
    'octa': OCTA,
    'local': LOCAL,
    '': LOCAL,

    'rdf': rdflib.RDF,
    'rdfs': rdflib.RDFS,
    'xsd': rdflib.XSD,
    'schema': rdflib.SDO,
    'sh': rdflib.SH,
    'skos': rdflib.SKOS,
    'owl': rdflib.OWL,
    'dc': rdflib.DC,
    'dcat': rdflib.DCAT,
    'dcterms': rdflib.DCTERMS,
    'foaf': rdflib.FOAF,
})


class Triple(NamedTuple):
    """RDF triple."""

    subject: rdflib.URIRef
    predicate: rdflib.URIRef
    object: Union[rdflib.URIRef, rdflib.Literal]  # noqa: WPS125

    def as_quad(self, graph: rdflib.URIRef) -> 'Quad':
        """Add graph to this triple and hence get a quad."""
        return Quad(self.subject, self.predicate, self.object, graph)


class Quad(NamedTuple):
    """Triple assigned to a named graph."""

    subject: rdflib.URIRef
    predicate: rdflib.URIRef
    object: Union[rdflib.URIRef, rdflib.Literal]  # noqa: WPS125
    graph: rdflib.URIRef


# Should be Dict[str, 'Context'] but mypy crashes on a recursive type.
Context = Optional[   # type: ignore
    Union[str, int, float, Dict[str, Any]],
]



import sys
from dataclasses import dataclass
from typing import Union

import rdflib
from mkdocs.structure.nav import Navigation, Section
from mkdocs.structure.pages import Page

from octadocs.conversions import get_page_title_by_iri, iri_by_page

if sys.version_info >= (3, 8):
    from functools import cached_property  # noqa
else:
    from backports.cached_property import cached_property  # noqa: WPS433,WPS440


NavigationItem = Union[Page, Section]


@dataclass(repr=False)
class OctadocsNavigationProcessor:
    """Rewrite navigation based on the knowledge graph."""

    graph: rdflib.ConjunctiveGraph
    navigation: Navigation

    @cached_property
    def homepage(self) -> Page:
        """Find the home page of the documentation site."""
        (home_page, ) = filter(
            lambda navigation_item: (
                isinstance(navigation_item, Page) and
                navigation_item.is_homepage
            ),
            self.navigation.items,
        )

        return home_page

    def generate(self) -> Navigation:
        """Generate the navigation structure."""
        navigation_items = list(map(
            self._process_nav_item,
            self.navigation.items,
        ))

        self.navigation.items = navigation_items

        return self.navigation

    def _process_nav_item(self, item: NavigationItem) -> NavigationItem:
        if isinstance(item, Page):
            return self._process_nav_page(item)

        elif isinstance(item, Section):
            return self._process_nav_section(item)

        else:
            raise Exception(f'What the heck is {item}?')

    def _process_nav_page(self, page: Page) -> Page:
        """Add title to a page."""
        iri = iri_by_page(page)

        title = get_page_title_by_iri(
            iri=iri,
            graph=self.graph,
        )

        if title:
            page.title = title

        return page

    def _process_nav_section(self, section: Section) -> NavigationItem:
        """Process section."""
        section.children = list(map(
            self._process_nav_item,
            section.children,
        ))

        if len(section.children) == 1:
            index_page = section.children[0]
            index_page.parent = section.parent
            return index_page

        index_page_candidates = list(filter(
            lambda page: page.file.src_path.endswith('index.md'),
            section.children,
        ))

        if index_page_candidates:
            index_page = index_page_candidates[0]
            section.title = index_page.title

            if section.parent is not None:
                index_page.parent = section.parent

        return section



from typing import Optional

import rdflib
from mkdocs.structure.pages import Page

from octadocs.environment import query, src_path_to_iri
from rdflib import URIRef


def iri_by_page(page: Page) -> URIRef:
    """Convert src_path of a file to a Zet IRI."""
    return src_path_to_iri(page.file.src_path)


def get_page_title_by_iri(
    iri: str,
    graph: rdflib.ConjunctiveGraph,
) -> Optional[str]:
    results = query(
        query_text='''
            SELECT ?title WHERE {
                ?page octa:title ?title .
            }
        ''',
        instance=graph,

        page=iri,
    )

    if results:
        return results[0]['title']

    return None



import logging
import operator
from functools import partial
from pathlib import Path
from typing import Callable, Dict, Optional, Union

import rdflib
from mkdocs.plugins import BasePlugin
from mkdocs.structure.files import Files
from mkdocs.structure.nav import Navigation, Section
from mkdocs.structure.pages import Page
from typing_extensions import TypedDict

from octadocs import settings
from octadocs.environment import query, src_path_to_iri
from octadocs.navigation import OctadocsNavigationProcessor
from octadocs.octiron import Octiron

NavigationItem = Union[Page, Section]

logger = logging.getLogger(__name__)


class ConfigExtra(TypedDict):
    """Extra portion of the config which we put our graph into."""

    graph: rdflib.ConjunctiveGraph


class Config(TypedDict):
    """MkDocs configuration."""

    docs_dir: str
    extra: Optional[ConfigExtra]


class TemplateContext(TypedDict):
    """Context for the native MkDocs page rendering engine."""

    graph: rdflib.ConjunctiveGraph
    iri: rdflib.URIRef
    this: rdflib.URIRef
    query: Callable[[str], Dict[str, rdflib.term.Identifier]]

    # FIXME this is hardcode and should be removed
    rdfs: rdflib.Namespace


def get_template_by_page(
    page: Page,
    graph: rdflib.ConjunctiveGraph,
) -> Optional[str]:
    """Find the template to render the given Markdown file."""
    iri = rdflib.URIRef(f'{settings.LOCAL_IRI_SCHEME}{page.file.src_path}')

    bindings = graph.query(
        'SELECT ?template_name WHERE { ?iri octa:template ?template_name }',
        initBindings={
            'iri': iri,
        },
    ).bindings

    if bindings:
        return bindings[0]['template_name'].value

    return None


class OctaDocsPlugin(BasePlugin):
    """MkDocs Meta plugin."""

    octiron: Octiron

    def on_config(self, config: Config) -> Config:
        """Initialize Octiron and provide graph to macros through the config."""
        self.octiron = Octiron(
            root_directory=Path(config['docs_dir']),
        )

        if config['extra'] is None:
            config['extra'] = {'graph': self.octiron.graph}

        else:
            config['extra']['graph'] = self.octiron.graph

        return config

    def on_files(self, files: Files, config: Config):
        """Extract metadata from files and compose the site graph."""
        for mkdocs_file in files:
            self.octiron.update_from_file(
                path=Path(mkdocs_file.abs_src_path),
                local_iri=src_path_to_iri(mkdocs_file.src_path),
                global_url=f'/{mkdocs_file.url}',
            )

        # After all files are imported, run inference rules.
        self.octiron.apply_inference()

    def on_page_markdown(
        self,
        markdown: str,
        page: Page,
        config: Config,
        files: Files,
    ):
        """Inject page template path, if necessary."""
        template_name = get_template_by_page(
            page=page,
            graph=self.octiron.graph,
        )

        if template_name is not None:
            page.meta['template'] = template_name

        return markdown

    def on_page_context(
        self,
        context: TemplateContext,
        page: Page,
        config: Config,
        nav: Page,
    ) -> TemplateContext:
        """Attach the views to certain pages."""
        page_iri = rdflib.URIRef(
            f'{settings.LOCAL_IRI_SCHEME}{page.file.src_path}',
        )

        this_choices = list(map(
            operator.itemgetter(rdflib.Variable('this')),
            self.octiron.graph.query(
                'SELECT * WHERE { ?this octa:subjectOf ?page_iri }',
                initBindings={
                    'page_iri': page_iri,
                },
            ).bindings,
        ))

        if this_choices:
            context['this'] = this_choices[0]
        else:
            context['this'] = page_iri

        context['graph'] = self.octiron.graph
        context['iri'] = page_iri

        context['query'] = partial(
            query,
            instance=self.octiron.graph,
        )
        # FIXME this is hardcode, needs to be defined dynamically
        context['rdfs'] = rdflib.Namespace(
            'http://www.w3.org/2000/01/rdf-schema#',
        )

        return context

    def on_nav(
        self,
        nav: Navigation,
        config: Config,
        files: Files,
    ) -> Navigation:
        """Update the site's navigation from the knowledge graph."""
        return OctadocsNavigationProcessor(
            graph=self.octiron.graph,
            navigation=nav,
        ).generate()



from typing import Any, Dict

from boltons.iterutils import remap

MetaData = Dict[str, Any]   # type: ignore


def convert_dollar_signs(
    meta_data: MetaData,
) -> MetaData:
    """
    Convert $ character to @ in keys.

    We use $ by convention to avoid writing quotes.
    """
    return remap(
        meta_data,
        lambda path, key, value: (  # noqa: WPS110
            key.replace('$', '@') if isinstance(key, str) else key,
            value,
        ),
    )



import json
from itertools import starmap
from typing import Iterator

import frontmatter
import rdflib
from pyld import jsonld
from rdflib import RDF

from octadocs.octiron.plugins import Loader
from octadocs.octiron.types import OCTA, Triple
from octadocs.octiron.yaml_extensions import convert_dollar_signs


class MarkdownLoader(Loader):
    """Load semantic data from Markdown front matter."""

    regex = r'\.md$'

    def stream(self) -> Iterator[Triple]:
        """Return stream of triples."""
        meta_data = frontmatter.load(self.path).metadata

        if not meta_data:
            return

        meta_data = convert_dollar_signs(meta_data)

        if meta_data.get('@context'):
            raise ValueError('A-A-A!!! @context is specified in front matter!')

        if meta_data.get('@id') is not None:
            # The author specified an IRI the document tells us about. Let us
            # link this IRI to the local document IRI.
            meta_data['octa:subjectOf'] = self.local_iri

        else:
            # The document author did not tell us about what their document is.
            # In this case, we assume that the local_iri of the document file
            # is the subject of the document description.
            meta_data['@id'] = self.local_iri

        # Reason: https://github.com/RDFLib/rdflib-jsonld/issues/97
        # If we don't expand with an explicit @base, import will fail silently.
        meta_data = jsonld.expand(
            meta_data,
            options={
                'base': 'local:',
                'expandContext': self.context,
            },
        )

        # Reason: https://github.com/RDFLib/rdflib-jsonld/issues/98
        # If we don't flatten, @included sections will not be imported.
        meta_data = jsonld.flatten(meta_data)

        serialized_meta_data = json.dumps(meta_data, indent=4)

        graph = rdflib.Graph()

        graph.parse(
            data=serialized_meta_data,
            format='json-ld',
        )

        # The IRI of the local page is a page.
        # FIXME: maybe this should be in octiron.py and work globally.
        yield Triple(self.local_iri, RDF.type, OCTA.Page)

        # The page will be available on the Web under certain URL.
        if self.global_url is not None:
            yield Triple(
                self.local_iri,
                OCTA.url,
                rdflib.Literal(self.global_url),
            )

        yield from starmap(Triple, iter(graph))



import json
import logging
import re
from dataclasses import dataclass, field
from functools import cached_property, partial, reduce
from pathlib import Path
from types import MappingProxyType
from typing import Dict, Iterable, Iterator, Optional, Type

import owlrl
import rdflib
import yaml
from deepmerge import always_merger

from octadocs.octiron.plugins import Loader, MarkdownLoader, TurtleLoader
from octadocs.octiron.types import DEFAULT_NAMESPACES, Context, Quad, Triple

logger = logging.getLogger(__name__)

CONTEXT_FORMATS = MappingProxyType({
    'context.json': json.load,

    # FIXME we need $ conversion for YAML files.
    'context.yaml': partial(
        yaml.load,
        Loader=Loader,
    ),
})

DEFAULT_CONTEXT = MappingProxyType({
    '@vocab': 'local:',
    '@base': 'local:',

    'rdfs': 'http://www.w3.org/2000/01/rdf-schema#',
    'schema': 'https://schema.org/',
    'octa': 'https://ns.octadocs.io/',

    'label': 'rdfs:label',
    'comment': 'rdfs:comment',
    'rdfs:isDefinedBy': {
        '@type': '@id',
    },
    'rdfs:subClassOf': {
        '@type': '@id',
    },
    'octa:subjectOf': {
        '@type': '@id',
    },
})


def triples_to_quads(
    triples: Iterator[Triple],
    graph: rdflib.URIRef,
) -> Iterator[Quad]:
    """Convert sequence of triples to sequence of quads."""
    yield from (
        triple.as_quad(graph)
        for triple in triples
    )


@dataclass
class Octiron:
    """Convert a lump of goo and data into a semantic graph."""

    root_directory: Path
    namespaces: Dict[str, rdflib.URIRef] = field(default_factory=dict)

    @cached_property
    def graph(self) -> rdflib.ConjunctiveGraph:
        """Generate and instantiate the RDFLib graph instance."""
        conjunctive_graph = rdflib.ConjunctiveGraph()

        namespaces = dict(DEFAULT_NAMESPACES)
        namespaces.update(self.namespaces)

        for short_name, uri in namespaces.items():
            conjunctive_graph.bind(short_name, uri)

        return conjunctive_graph

    def get_context_per_directory(
        self,
        directory: Path,
    ) -> Context:
        """Find context file per disk directory."""
        return reduce(
            always_merger.merge,
            map(
                self._get_context_file,
                reversed(list(self._find_context_files(directory))),
            ),
            dict(DEFAULT_CONTEXT),
        )

    def update_from_file(
        self,
        path: Path,
        local_iri: rdflib.URIRef,
        global_url: Optional[str] = None,
    ) -> None:
        """Update the graph from file determined by given path."""
        context = self.get_context_per_directory(path.parent)
        loader_class = self.get_loader_class_for_path(path)

        if loader_class is None:
            return

        loader_instance = loader_class(
            path=path,
            context=context,
            local_iri=local_iri,
            global_url=global_url,
        )
        triples = loader_instance.stream()

        quads = triples_to_quads(triples=triples, graph=local_iri)

        self.graph.addN(quads)

    def apply_inference(self) -> None:
        """Do whatever is needed after the graph was updated from a file."""
        # FIXME this should be customizable via dependency inversion. Right now
        #   this is hardcoded to run inference rules formulated as SPARQL files.
        logger.info('Inference: OWL RL')
        owlrl.DeductiveClosure(owlrl.OWLRL_Extension).expand(self.graph)

        # Fill in octa:about relationships.
        logger.info(
            'Inference: '
            '?thing octa:subjectOf ?page ⇒ ?page octa:about ?thing .',
        )
        self.graph.update('''
            INSERT {
                ?page octa:about ?thing .
            } WHERE {
                ?thing octa:subjectOf ?page .
            }
        ''')

        logger.info(
            'Inference: ?thing rdfs:label ?label & '
            '?thing octa:page ?page ⇒ ?page octa:title ?label',
        )
        self.graph.update('''
            INSERT {
                ?page octa:title ?title .
            } WHERE {
                ?subject
                    rdfs:label ?title ;
                    octa:subjectOf ?page .
            }
        ''')

        inference_dir = self.root_directory.parent / 'inference'
        if inference_dir.is_dir():
            for sparql_file in inference_dir.iterdir():
                logger.info('Inference: %s', sparql_file.name)
                sparql_text = sparql_file.read_text()
                self.graph.update(sparql_text)

    def get_loader_class_for_path(self, path: Path) -> Optional[Type[Loader]]:
        """Based on file path, determine the loader to use."""
        # TODO dependency inversion
        loaders = [MarkdownLoader, TurtleLoader]

        for loader in loaders:
            if re.search(loader.regex, str(path)) is not None:
                return loader

        logger.warning('Cannot find appropriate loader for path: %s', path)
        return None

    def _find_context_files(self, directory: Path) -> Iterable[Path]:
        """
        Find all context files relevant to particular directory.

        Files are ordered from the deepest to the upmost.
        """
        for context_directory in (directory, *directory.parents):
            for filename in CONTEXT_FORMATS.keys():
                context_path = context_directory / filename

                if context_path.is_file():
                    yield context_path

            if context_directory == self.root_directory:
                return

    def _get_context_file(self, path: Path) -> Context:
        """Read and return context file by path."""
        loader = CONTEXT_FORMATS[path.name]

        with path.open('r') as context_data_stream:
            return loader(context_data_stream)  # type: ignore



import re
from typing import Dict, List

import rdflib
from rdflib import URIRef, term
from rdflib.plugins.sparql.processor import SPARQLResult

from octadocs.settings import LOCAL_IRI_SCHEME


def query(
    query_text: str,
    instance: rdflib.ConjunctiveGraph,
    **kwargs: str,
):
    """Run SPARQL SELECT query and return formatted result."""
    sparql_result: SPARQLResult = instance.query(
        query_text,
        initBindings=kwargs,
    )

    if sparql_result.askAnswer is not None:
        return sparql_result.askAnswer

    return _format_query_bindings(sparql_result.bindings)


def iri_to_url(iri: str) -> str:
    """Convert Zet IRI into clickable URL."""
    iri = iri.replace(LOCAL_IRI_SCHEME, '')

    if iri.endswith('index.md'):
        return re.sub(
            r'/index\.md$',
            '/',
            iri,
        )

    else:
        return re.sub(
            r'\.md$',
            '',
            iri,
        )


def src_path_to_iri(src_path: str) -> URIRef:
    """Convert src_path of a file to a Zet IRI."""
    return URIRef(f'{LOCAL_IRI_SCHEME}{src_path}')


def _format_sparql_variable_value(rdf_value: term.Identifier) -> str:
    """Format a RDF value as a primitive type value."""
    if isinstance(rdf_value, rdflib.URIRef):
        return str(rdf_value)

    elif isinstance(rdf_value, rdflib.BNode):
        return str(rdf_value)

    elif isinstance(rdf_value, rdflib.Literal):
        return rdf_value.value

    raise ValueError(f'Cannot interpret value: {rdf_value}')


def _format_query_bindings(
    bindings: List[Dict[rdflib.Variable, term.Identifier]],
) -> List[Dict[str, str]]:
    """
    Format bindings before returning them.

    FIXME: I am not sure this is really necessary.
    """
    return [
        {
            str(variable_name): _format_sparql_variable_value(rdf_value)
            for variable_name, rdf_value
            in row.items()
        }
        for row in bindings
    ]



import html
import io
from base64 import b64encode
from functools import partial
from typing import Any, Dict, Optional
from unittest.mock import patch

import pydotplus
import rdflib
from macros.plugin import MacrosPlugin
from rdflib import Variable
from rdflib.plugins.sparql.processor import SPARQLResult
from rdflib.tools.rdf2dot import rdf2dot

from octadocs.conversions import iri_by_page
from octadocs.environment import iri_to_url, query, src_path_to_iri


def graph(instance: rdflib.ConjunctiveGraph) -> str:
    """
    Render RDF graph visually as PNG image.

    Idea: https://stackoverflow.com/a/61483971/1245471
    """
    dot_description = io.StringIO()

    with patch('rdflib.tools.rdf2dot.cgi', html):
        # FIXME hack, fixes this: https://github.com/RDFLib/rdflib/issues/1110
        rdf2dot(instance, dot_description)

    dg = pydotplus.graph_from_dot_data(dot_description.getvalue())
    png = dg.create_png()

    encoded_png = b64encode(png).decode('utf-8')

    return f'<img src="data:image/png;base64,{encoded_png}" />'


def turtle(instance: rdflib.ConjunctiveGraph) -> str:
    """Serialize graph as n3."""
    serialized_document = instance.serialize(format='turtle').decode('utf-8')
    return (
        '```n3\n'
        f'{serialized_document}\n'
        '```\n'
    )


def sparql(
    instance: rdflib.ConjunctiveGraph,
    query: str,
    **kwargs: str,
) -> SPARQLResult:
    bindings = {
        argument_name: argument_value
        for argument_name, argument_value in kwargs.items()
    }

    return instance.query(query, initBindings=bindings)


def _render_as_row(row: Dict[Variable, Any]) -> str:  # type: ignore
    """Render row of a Markdown table."""
    formatted_row = ' | '.join(row.values())
    return f'| {formatted_row} |'


def table(query_result: SPARQLResult) -> str:
    """Render as a Markdown table."""
    headers = ' | '.join(str(cell) for cell in query_result.vars)

    rows = '\n'.join(
        _render_as_row(row)
        for row in query_result.bindings
    )

    separators = '| ' + (' --- |' * len(query_result.vars))  # noqa: WPS336

    return f'''
---
| {headers} |
{separators}
{rows}
'''


def construct(
    query_text: str,
    instance: rdflib.ConjunctiveGraph,
    **kwargs: str,
) -> rdflib.Graph:
    """Run SPARQL SELECT query and return formatted result."""
    sparql_result: SPARQLResult = instance.query(
        query_text,
        initBindings=kwargs,
    )

    return sparql_result.graph


def url(
    resource: rdflib.URIRef,
    graph: rdflib.ConjunctiveGraph
) -> Optional[str]:
    """Convert a URIRef to a clickable URL."""
    bindings = graph.query(
        'SELECT ?url WHERE { ?resource octa:subjectOf/octa:url ?url . } ',
        initBindings={
            'resource': resource,
        }
    ).bindings

    if not bindings:
        return None

    return '/' + bindings[0][rdflib.Variable('url')].value


def label(
    resource: rdflib.URIRef,
    graph: rdflib.ConjunctiveGraph
) -> Optional[str]:
    """Convert a URIRef to a clickable URL."""
    bindings = graph.query(
        'SELECT ?label WHERE { ?resource rdfs:label ?label . } ',
        initBindings={
            'resource': resource,
        }
    ).bindings

    if not bindings:
        return None

    return bindings[0][rdflib.Variable('label')].value


def define_env(env: MacrosPlugin) -> MacrosPlugin:  # noqa: WPS213
    """Create mkdocs-macros Jinja environment."""
    env.filter(graph)
    env.filter(sparql)
    env.filter(turtle)
    env.filter(table)

    env.macro(
        partial(
            query,
            instance=env.variables.graph,
        ),
        name='query',
    )

    env.macro(
        partial(
            construct,
            instance=env.variables.graph,
        ),
        name='construct',
    )

    env.macro(
        partial(
            url,
            graph=env.variables.graph,
        ),
        name='url',
    )

    env.filter(
        partial(
            url,
            graph=env.variables.graph,
        ), name='url',
    )

    env.filter(
        partial(
            label,
            graph=env.variables.graph,
        ),
        name='label',
    )

    env.macro(iri_to_url)
    env.macro(src_path_to_iri)

    env.filter(iri_to_url)

    env.filter(iri_by_page)
    env.macro(iri_by_page)

    # FIXME this is hardcode, needs to be defined dynamically
    env.variables['rdfs'] = rdflib.Namespace(
        'http://www.w3.org/2000/01/rdf-schema#',
    )
    env.variables['local'] = rdflib.Namespace('local:')

    env.variables['URIRef'] = rdflib.URIRef

    return env
